use std::{fmt::Display, iter::repeat};

use itertools::Itertools;

enum IncrementResult {
    Overflow,
    Good,
}
use IncrementResult::*;

trait Incrementable {
    fn increment(&mut self) -> IncrementResult;
}

impl Incrementable for Vec<bool> {
    fn increment(&mut self) -> IncrementResult {
        for a in self {
            *a = !*a;
            if *a {
                return Good;
            }
        }
        return Overflow;
    }
}

#[derive(Debug, Default, Clone)]
struct Atom {
    smaller: Vec<bool>,
}

impl Incrementable for Atom {
    fn increment(&mut self) -> IncrementResult {
        self.smaller.increment()
    }
}

#[derive(Debug, Default, Clone)]
struct AtomSet {
    included: Vec<bool>,
}

impl Incrementable for AtomSet {
    fn increment(&mut self) -> IncrementResult {
        self.included.increment()
    }
}

#[derive(Debug, Default)]
struct Lattice<I: Incrementable + Default, Ext> {
    ext: Ext,
    order: Vec<I>,
}

trait LatticeAbstract {
    fn le(&self, a: usize, b: usize) -> bool;
    fn grow(&mut self) -> IncrementResult;
    fn is_lattice(&self) -> bool;
}

impl Display for Lattice<Atom, ()> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for (i, atom) in self.order.iter().enumerate() {
            let s = atom
                .smaller
                .iter()
                .map(|&b| if b { "1 " } else { "0 " })
                .collect::<String>();
            writeln!(f, "{}: {}", i, s)?;
        }
        Ok(())
    }
}

impl<I: Incrementable + Default, Ext> Incrementable for Lattice<I, Ext> {
    fn increment(&mut self) -> IncrementResult {
        for element in &mut self.order {
            if let Good = element.increment() {
                return Good;
            }
        }
        return Overflow;
    }
}

impl<I: Incrementable + Default, Ext> Lattice<I, Ext>
where
    Self: LatticeAbstract,
{
    fn next_lattice(&mut self) -> IncrementResult {
        loop {
            if let Overflow = self.increment() {
                if let Overflow = self.grow() {
                    return Overflow;
                }
            }
            if self.is_lattice() {
                return Good;
            }
        }
    }
    fn ge(&self, a: usize, b: usize) -> bool {
        return self.le(b, a);
    }
    fn is_bounded<C, F>(&self, subset: &Vec<usize>, bound_candidates: C, rel: F) -> bool
    where
        C: Iterator<Item = usize>,
        F: Fn(&Self, usize, usize) -> bool,
    {
        let mut bound: Option<usize> = None;
        for new_bound in bound_candidates {
            let is_bound = || subset.iter().all(|&el| rel(self, el, new_bound));
            if let Some(old_bound) = bound {
                if rel(self, old_bound, new_bound) {
                    continue;
                } else if rel(self, new_bound, old_bound) {
                    panic!("Bound candidates must be sorted topologically");
                } else if is_bound() {
                    return false;
                }
            } else if is_bound() {
                bound = Some(new_bound);
            }
        }
        return true;
    }

    fn has_lub_glb(&self) -> bool {
        // A top/bottom element are inherent as are partial order conditions.
        // Lattice condition can only fail if a set of elements have non comparable upper/lower bounds
        // Possible optimization: a lack of LUB might imply no GLB so may only need to check one of them
        let mut subsets = (0..self.order.len()).powerset();
        subsets.next(); // Skip the empty set
        for subset in subsets {
            if !self.is_bounded(&subset, 0..self.order.len(), Lattice::le) {
                return false;
            }
            if !self.is_bounded(&subset, (0..self.order.len()).rev(), Lattice::ge) {
                return false;
            }
        }
        return true;
    }
    fn order_preserving<IncB, ExtB>(&self, other: &Lattice<IncB, ExtB>) -> bool
    where
        IncB: Incrementable + Default,
        Lattice<IncB, ExtB>: LatticeAbstract,
    {
        assert_eq!(self.order.len(), other.order.len());
        for (a, b) in (0..self.order.len()).cartesian_product(0..other.order.len()) {
            if self.le(a, b) && !other.le(a, b) {
                return false;
            }
        }
        return true;
    }
    fn equal<IncB, ExtB>(&self, other: &Lattice<IncB, ExtB>) -> bool
    where
        IncB: Incrementable + Default,
        Lattice<IncB, ExtB>: LatticeAbstract,
    {
        self.order_preserving(other) && other.order_preserving(self)
    }
}

impl Lattice<Atom, ()> {
    fn new() -> Self {
        Self::default()
    }
    fn is_transitive(&self) -> bool {
        for (i, j) in (0..self.order.len()).cartesian_product(0..self.order.len()) {
            if !self.le(i, j) && self.le_transitive(i, j) {
                return false;
            }
        }
        return true;
    }
    // The underlying ordering might not be transitive, this makes it so
    fn le_transitive(&self, a: usize, b: usize) -> bool {
        if self.le(a, b) {
            return true;
        }
        for (b, &active) in self.order[b].smaller.iter().enumerate() {
            if active && self.le_transitive(a, b) {
                return true;
            }
        }
        return false;
    }
}

impl LatticeAbstract for Lattice<Atom, ()> {
    fn le(&self, a: usize, b: usize) -> bool {
        if a == b {
            return true;
        }
        if a > b {
            return false;
        }
        if self.order[b].smaller[a] {
            return true;
        }
        return false;
    }
    fn grow(&mut self) -> IncrementResult {
        self.order = vec![Default::default(); self.order.len() + 1];
        for (i, atom) in self.order.iter_mut().enumerate() {
            *atom = Atom {
                smaller: vec![false; i],
            }
        }
        return Good;
    }
    fn is_lattice(&self) -> bool {
        self.is_transitive() && self.has_lub_glb()
    }
}

#[cfg(test)]
mod lattice_tests {
    use super::*;
    use std::error::Error;

    #[test]
    fn test_is_lattice() -> Result<(), Box<dyn Error>> {
        let lattice = Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![true],
                },
                Atom {
                    smaller: vec![true, false],
                },
            ],
        };
        assert!(lattice.is_lattice());
        let lattice = Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![false],
                },
                Atom {
                    smaller: vec![true, true],
                },
            ],
        };
        assert!(lattice.is_lattice());
        Ok(())
    }
    #[test]
    fn test_is_not_lattice() -> Result<(), Box<dyn Error>> {
        let lattice = Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![true],
                },
                Atom {
                    smaller: vec![false, true],
                },
            ],
        };
        assert!(!lattice.is_transitive());
        let lattice = Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![false],
                },
                Atom {
                    smaller: vec![true, true],
                },
                Atom {
                    smaller: vec![true, true, false],
                },
            ],
        };
        assert!(lattice.is_transitive());
        assert!(!lattice.is_lattice());
        Ok(())
    }
}

struct Mapping<'a> {
    domain: &'a Lattice<Atom, ()>,
    image: &'a Lattice<Atom, ()>,
}

// function: Lattice<AtomSet>,

type Operator<'a> = Lattice<AtomSet, Mapping<'a>>;

impl<'a> Operator<'a> {
    fn new(domain: &'a Lattice<Atom, ()>, image: &'a Lattice<Atom, ()>) -> Operator<'a> {
        Lattice {
            ext: Mapping { domain, image },
            order: vec![
                AtomSet {
                    included: vec![false; domain.order.len()]
                };
                image.order.len()
            ],
        }
    }
    fn all_nonempty(&self) -> bool {
        self.order.iter().all(|set| set.included.iter().any(|&x| x))
    }
    fn le_helper<I, F>(&self, a: I, b: I, rel: F) -> bool
    where
        I: Iterator<Item = usize> + Clone,
        F: Fn(&Lattice<Atom, ()>, usize, usize) -> bool,
    {
        b.clone()
            .all(|right| a.clone().any(|left| rel(&self.ext.domain, left, right)))
    }
}

impl<'a> LatticeAbstract for Operator<'a> {
    fn le(&self, a: usize, b: usize) -> bool {
        let pick = |which: usize| {
            self.order[which]
                .included
                .iter()
                .enumerate()
                .filter_map(|(i, &included)| if included { Some(i) } else { None })
        };
        let left = pick(a);
        let right = pick(b);

        if !self.le_helper(left.clone(), right.clone(), Lattice::<Atom, ()>::le) {
            return false;
        }
        if !self.le_helper(right, left, Lattice::<Atom, ()>::ge) {
            return false;
        }
        return true;
    }

    fn grow(&mut self) -> IncrementResult {
        return Overflow;
    }

    fn is_lattice(&self) -> bool {
        self.all_nonempty() && self.ext.image.equal(self)
    }
}

impl<'a> Display for Operator<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for (i, atom) in self.order.iter().enumerate() {
            let s = atom
                .included
                .iter()
                .map(|&b| if b { "1 " } else { "0 " })
                .collect::<String>();
            writeln!(f, "{}: {}", i, s)?;
        }
        Ok(())
    }
}

#[cfg(test)]
mod operator_tests {
    use super::*;
    use std::error::Error;

    #[test]
    fn test_operator() {
        let domain = &Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![true],
                },
                Atom {
                    smaller: vec![true, false],
                },
            ],
        };
        let image = &Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![true],
                },
                Atom {
                    smaller: vec![true, false],
                },
            ],
        };
        assert!(domain.is_lattice());
        assert!(image.is_lattice());
        assert!(image.equal(domain));
        // Identity operator
        let operator = Lattice {
            ext: Mapping { domain, image },
            order: vec![
                AtomSet {
                    included: vec![true, false, false],
                },
                AtomSet {
                    included: vec![false, true, false],
                },
                AtomSet {
                    included: vec![false, false, true],
                },
            ],
        };
        assert!(operator.is_lattice());
    }

    #[test]
    fn test_order_preserving() {
        let lattice_a = Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![true],
                },
                Atom {
                    smaller: vec![true, false],
                },
            ],
        };
        let lattice_b = Lattice {
            ext: (),
            order: vec![
                Atom { smaller: vec![] },
                Atom {
                    smaller: vec![true],
                },
                Atom {
                    smaller: vec![true, false],
                },
            ],
        };
        assert!(lattice_a.order_preserving(&lattice_b));
    }
}

fn main() {
    let mut domain = Lattice::<Atom, ()>::new();
    let image = Lattice {
        ext: (),
        order: vec![
            Atom { smaller: vec![] },
            Atom {
                smaller: vec![true],
            },
            Atom {
                smaller: vec![true, false],
            },
        ],
    };
    while let Good = domain.next_lattice() {
        if domain.order.len() != image.order.len() || !domain.order_preserving(&image) {
            continue;
        }
        println!("{}", domain);
        let mut operator = Operator::new(&domain, &image);
        while let Good = operator.increment() {
            // println!("ah: {}", operator);
            // TODO doesn't work because we generate a domain larger than the image lol
        }
    }
}
